@@ -274,7 +274,7 @@ type EBNF struct {
 	Productions []*Production `{ @@ }`
 }
 
-func TestEBNF(t *testing.T) {
+func TestEBNFParser(t *testing.T) {
 	parser := mustTestParser(t, &EBNF{})
 
 	expected := &EBNF{
@@ -560,13 +560,16 @@ func TestRepeatAcrossFields(t *testing.T) {
 
 func TestPosInjection(t *testing.T) {
 	type subgrammar struct {
-		Pos lexer.Position
-		B   string `@{ "," }`
+		Pos    lexer.Position
+		B      string `@{ "," }`
+		EndPos lexer.Position
 	}
 	type grammar struct {
-		Pos lexer.Position
-		A   string      `@{ "." }`
-		B   *subgrammar `@@`
+		Pos    lexer.Position
+		A      string      `@{ "." }`
+		B      *subgrammar `@@`
+		C      string      `@"."`
+		EndPos lexer.Position
 	}
 
 	parser := mustTestParser(t, &grammar{})
@@ -586,10 +589,21 @@ func TestPosInjection(t *testing.T) {
 				Line:   1,
 				Column: 7,
 			},
+			EndPos: lexer.Position{
+				Offset: 9,
+				Line:   1,
+				Column: 10,
+			},
+		},
+		C: ".",
+		EndPos: lexer.Position{
+			Offset: 10,
+			Line:   1,
+			Column: 11,
 		},
 	}
 
-	err := parser.ParseString("   ...,,,", actual)
+	err := parser.ParseString("   ...,,,.", actual)
 	require.NoError(t, err)
 	require.Equal(t, expected, actual)
 }
@@ -614,6 +628,26 @@ func TestCaptureInterface(t *testing.T) {
 	require.Equal(t, expected, actual)
 }
 
+type unmarshallableCount int
+
+func (u *unmarshallableCount) UnmarshalText(text []byte) error {
+	*u += unmarshallableCount(len(text))
+	return nil
+}
+
+func TestTextUnmarshalerInterface(t *testing.T) {
+	type grammar struct {
+		Count unmarshallableCount `{ @"a" }`
+	}
+
+	parser := mustTestParser(t, &grammar{})
+	actual := &grammar{}
+	expected := &grammar{Count: 3}
+	err := parser.ParseString("a a a", actual)
+	require.NoError(t, err)
+	require.Equal(t, expected, actual)
+}
+
 func TestLiteralTypeConstraint(t *testing.T) {
 	type grammar struct {
 		Literal string `@"123456":String`
@@ -659,7 +693,7 @@ type parseableStruct struct {
 	Tokens []string
 }
 
-func (p *parseableStruct) Parse(lex lexer.PeekingLexer) error {
+func (p *parseableStruct) Parse(lex *lexer.PeekingLexer) error {
 	tokens, err := lexer.ConsumeAll(lex)
 	if err != nil {
 		return err
@@ -1042,7 +1076,7 @@ type Issue62Bar struct {
 	A int
 }
 
-func (x *Issue62Bar) Parse(lex lexer.PeekingLexer) error {
+func (x *Issue62Bar) Parse(lex *lexer.PeekingLexer) error {
 	token, err := lex.Next()
 	if err != nil {
 		return err
@@ -1059,3 +1093,176 @@ func TestIssue62(t *testing.T) {
 	_, err := Build(&Issue62Foo{})
 	require.NoError(t, err)
 }
+
+// nolint: structcheck
+func TestIssue71(t *testing.T) {
+	type Sub struct {
+		name string `@Ident`
+	}
+	type grammar struct {
+		pattern *Sub `@@`
+	}
+
+	_, err := Build(&grammar{})
+	require.Error(t, err)
+}
+
+func TestAllowTrailing(t *testing.T) {
+	type G struct {
+		Name string `@Ident`
+	}
+
+	p, err := Build(&G{})
+	require.NoError(t, err)
+
+	g := &G{}
+	err = p.ParseString(`hello world`, g)
+	require.Error(t, err)
+	err = p.ParseString(`hello world`, g, AllowTrailing(true))
+	require.NoError(t, err)
+	require.Equal(t, &G{"hello"}, g)
+}
+
+func TestDisjunctionErrorReporting(t *testing.T) {
+	type statement struct {
+		Add    bool `  @"add"`
+		Remove bool `| @"remove"`
+	}
+	type grammar struct {
+		Statements []*statement `"{" ( @@ )* "}"`
+	}
+	p := mustTestParser(t, &grammar{})
+	ast := &grammar{}
+	err := p.ParseString(`{ add foo }`, ast)
+	// TODO: This should produce a more useful error. This is returned by sequence.Parse().
+	require.EqualError(t, err, `1:7: unexpected token "foo" (expected "}")`)
+}
+
+func TestCustomInt(t *testing.T) {
+	type MyInt int
+	type G struct {
+		Value MyInt `@Int`
+	}
+
+	p, err := Build(&G{})
+	require.NoError(t, err)
+
+	g := &G{}
+	err = p.ParseString(`42`, g)
+	require.NoError(t, err)
+	require.Equal(t, &G{42}, g)
+}
+
+func TestBoolIfSet(t *testing.T) {
+	type G struct {
+		Value bool `@"true"?`
+	}
+
+	p, err := Build(&G{})
+	require.NoError(t, err)
+
+	g := &G{}
+	err = p.ParseString(`true`, g)
+	require.NoError(t, err)
+	require.Equal(t, &G{true}, g)
+	err = p.ParseString(``, g)
+	require.NoError(t, err)
+	require.Equal(t, &G{false}, g)
+}
+
+func TestCustomBoolIfSet(t *testing.T) {
+	type MyBool bool
+	type G struct {
+		Value MyBool `@"true"?`
+	}
+
+	p, err := Build(&G{})
+	require.NoError(t, err)
+
+	g := &G{}
+	err = p.ParseString(`true`, g)
+	require.NoError(t, err)
+	require.Equal(t, &G{true}, g)
+	err = p.ParseString(``, g)
+	require.NoError(t, err)
+	require.Equal(t, &G{false}, g)
+}
+
+func TestPointerToList(t *testing.T) {
+	type grammar struct {
+		List *[]string `@Ident*`
+	}
+	p := mustTestParser(t, &grammar{})
+	ast := &grammar{}
+	err := p.ParseString(`foo bar`, ast)
+	require.NoError(t, err)
+	l := []string{"foo", "bar"}
+	require.Equal(t, &grammar{List: &l}, ast)
+}
+
+// I'm not sure if this is a problem that should be solved like this.
+
+// func TestMatchHydratesNullFields(t *testing.T) {
+// 	type grammar struct {
+// 		List []string `"{" @Ident* "}"`
+// 	}
+// 	p := mustTestParser(t, &grammar{})
+// 	ast := &grammar{}
+// 	err := p.ParseString(`{}`, ast)
+// 	require.NoError(t, err)
+// 	require.NotNil(t, ast.List)
+// }
+
+func TestNegation(t *testing.T) {
+	type grammar struct {
+		EverythingUntilSemicolon *[]string `@!';'* @';'`
+	}
+	p := mustTestParser(t, &grammar{})
+	ast := &grammar{}
+	err := p.ParseString(`hello world ;`, ast)
+	require.NoError(t, err)
+	require.Equal(t, &[]string{"hello", "world", ";"}, ast.EverythingUntilSemicolon)
+
+	err = p.ParseString(`hello world`, ast)
+	require.Error(t, err)
+}
+
+func TestNegationWithPattern(t *testing.T) {
+	type grammar struct {
+		EverythingMoreComplex *[]string `@!(';' String)* @';' @String`
+	}
+
+	p := mustTestParser(t, &grammar{})
+	// j, err := json.MarshalIndent(p.root, "", "  ")
+	// log.Print(j)
+	// log.Print(stringer(p.root))
+	ast := &grammar{}
+	err := p.ParseString(`hello world ; 'some-str'`, ast)
+	require.NoError(t, err)
+	require.Equal(t, &[]string{"hello", "world", ";", `some-str`}, ast.EverythingMoreComplex)
+
+	err = p.ParseString(`hello ; world ; 'hey'`, ast)
+	require.NoError(t, err)
+	require.Equal(t, &[]string{"hello", ";", "world", ";", `hey`}, ast.EverythingMoreComplex)
+
+	err = p.ParseString(`hello ; world ;`, ast)
+	require.Error(t, err)
+}
+
+func TestNegationWithDisjunction(t *testing.T) {
+	type grammar struct {
+		EverythingMoreComplex *[]string `@!(';' | ',')* @(';' | ',')`
+	}
+
+	// Note: we need more lookahead since (';' String) needs some before failing to match
+	p := mustTestParser(t, &grammar{})
+	ast := &grammar{}
+	err := p.ParseString(`hello world ;`, ast)
+	require.NoError(t, err)
+	require.Equal(t, &[]string{"hello", "world", ";"}, ast.EverythingMoreComplex)
+
+	err = p.ParseString(`hello world , `, ast)
+	require.NoError(t, err)
+	require.Equal(t, &[]string{"hello", "world", ","}, ast.EverythingMoreComplex)
+
+}
